package org.pd.parser;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.*;
import org.pd.parser.SourceFile;

public abstract class RuleReader {

	private File defFile;
	private File inDir;
	private String currentFile;

	private HashMap<String, String> variables;
	private HashMap<String, SourceFile> sources;

	private static Pattern varPattern = Pattern.compile("%[\\w.]+?%");

	/**
	 * @param defFile File object for the definition file
	 * @param inDir File object for the input directory
	 */
	public RuleReader(File defFile, File inDir) {
		if (!defFile.exists())
			throw new RuntimeException("Invalid definition file: " + defFile);
		if (!inDir.exists() || !inDir.isDirectory())
			throw new RuntimeException("Invalid input directory: " + inDir);

		this.defFile = defFile;
		this.inDir = inDir;
		this.variables = variables;

		variables = new HashMap<String, String>();
		sources = new HashMap<String, SourceFile>();
	}

	/**
	 * Simply calls <code>parse(defFile)</code>
	 */
	public void parse() {
		parse(defFile);
	}

	/**
	 * Reads the given rule file and delegates to the handling methods.
	 *
	 * @param f The file to parse
	 */
	public void parse(File f) {
		String rawLine = null;
		try {
			BufferedReader reader = new BufferedReader(new FileReader(f));
			String line;
			
			while ((line = reader.readLine()) != null) {
				if (line.trim().length() == 0 || line.startsWith("#"))
					continue;
				rawLine = line;
				line = expandVariables(line.trim());

				if (line.startsWith("file"))
					setCurrentFile(line.split(" ")[1]);
				else if (line.startsWith("include"))
					parse(new File(f.getParent(), line.split(" ")[1]));
				else if (line.startsWith("define"))
					variables.put(line.split(" ", 3)[1], line.split(" ", 3)[2]);
				else if (line.startsWith("declare"))
					addDeclaration(line.split(" ", 2)[1]);
				else if (line.startsWith("regex"))
					parseRegex(line.split(" ", 2)[1], reader, true);
				else if (line.startsWith("optregex"))
					parseRegex(line.split(" ", 2)[1], reader, false);
				else if (line.startsWith("inmethod"))
					inMethod(line.split(" ", 2)[1], reader, true);
				else if (line.startsWith("optinmethod"))
					inMethod(line.split(" ", 2)[1], reader, false);
				else
					throw new RuntimeException("Invalid beginning of rule: " + line);
			}
		} catch (IOException ioe) {
			throw new RuntimeException(ioe.toString());
		} catch (RuntimeException rte) {
			System.err.println("line = " + rawLine.trim());
			System.err.println("file = " + f.getAbsolutePath());
			throw rte;
		}
	}

	/**
	 * Parses the '(opt)regex' blocks.
	 *
	 * Delegates most work to <code>handleLine()</code>, which can be overridden.
	 * This simply loops the array returned from <code>getCurrentSources()</code>
	 * and attempts to match the pattern in each SourceFile. Does not
	 * continue to search once a match has been found.
	 *
	 * @param pattern The pattern to match compile
	 * @param reader The BufferedReader for the rule file
	 * @param fatal Whether a failure to match should be considered fatal
	 * @see #handleLine(String, Matcher)
	 * @see #getCurrentSources()
	 */
	private void parseRegex(String pattern, BufferedReader reader, boolean fatal) {
		Pattern p = Pattern.compile(pattern);
		SourceFile[] sfa = getCurrentSources();

		for (int i = 0; i < sfa.length; i++) {
			SourceFile sf = sfa[i];
			Matcher m = p.matcher(sf.getContents());
			boolean ignore = false;

			if (m.groupCount() == 0)
				throw new RuntimeException("Useless pattern; has no groups: " + pattern);
			if (!m.find()) {
				if (fatal && i == sfa.length - 1)
					throw new RuntimeException("Expression failed to match in file "
							+ sf.getName() + ": " + p );
				else if (i < sfa.length - 1)
					continue;
				else
					ignore = true;
			}

			try {
				String line;

				while ((line = reader.readLine()) != null) {
					if (line.trim().length() == 0 || line.startsWith("#"))
						continue;

					if (line.toLowerCase().startsWith("endrule"))
						break;

					if (!ignore)
						handleLine(line, m);
				}

				break;
			} catch (IOException ioe) {
				throw new RuntimeException(ioe.toString());
			}
		}
	}
	

	/**
	 * Parses the '(opt)inmethod' blocks.
	 *
	 * Similar to <code>parseRegex()</code>, this parses a single block after the
	 * '(opt)inmethod' keywords. It matches the pattern in the same
	 * manner as <code>parseRegex()</code>, but then uses <code>findContainerMethod()</code>
	 * to determine which method the match is inside of, and calls
	 * <code>handleExtraction()</code> directly.
	 *
	 * @param pattern The pattern to match compile
	 * @param reader The BufferedReader for the rule file
	 * @param fatal Whether a failure to match should be considered fatal
	 * @see #parseRegex(String, BufferedReader, boolean)
	 * @see #handleExtraction(String, String, String)
	 */
	public void inMethod(String pattern, BufferedReader reader, boolean fatal) {
		Pattern p = Pattern.compile(pattern);
		SourceFile[] sfa = getCurrentSources();

		for (int i = 0; i < sfa.length; i++) {
			SourceFile sf = sfa[i];
			Matcher m = p.matcher(sf.getContents());
			boolean ignore = false;

			if (!m.find()) {
				if (fatal && i == sfa.length - 1)
					throw new RuntimeException("Expression failed to match in file "
							+ sf.getName() + ": " + p);
				else if (i < sfa.length - 1)
					continue;
				else
					ignore = true;
			}

			try {
				String line;

				while ((line = reader.readLine()) != null) {
					if (line.trim().length() == 0 || line.startsWith("#"))
						continue;

					if (line.toLowerCase().startsWith("endrule"))
						break;

					if (!ignore) {
						String[] pieces = line.split(" ");
						if (pieces.length != 2)
							throw new RuntimeException("Invalid inmethod assignment: " + line);

						String methodName = findContainerMethod(sf, m);
						handleExtraction(pieces[0], pieces[1], methodName);
					}
				}

				break;
			} catch (IOException ioe) {
				throw new RuntimeException(ioe.toString());
			}
		}
	}

	/**
	 * Parses a single extraction instruction.
	 *
	 * Breaks the line into three pieces, the first two of which
	 * correspond to the first two arguments to <code>handleExtraction()</code>.
	 * The final piece should be an integer representing which of the
	 * pattern's groups to pass as the final argument to <code>handleExtraction()</code>.
	 *
	 * @param line The extraction instruction String
	 * @param m The matcher to use for groups
	 * @see #handleExtraction(String, String, String)
	 */
	public void handleLine(String line, Matcher m) {
		String[] pieces = line.split(" ");
		if (pieces.length != 3)
			throw new RuntimeException("Invalid extraction instruction: " + line);

		int matchIndex = -1;
		try {
			matchIndex = Integer.parseInt(pieces[2]);
		} catch (Exception e) {
			throw new RuntimeException("Couldn't parse integer " + pieces[2] + ": " + e);
		}

		if (matchIndex < 1 || matchIndex > m.groupCount())
			throw new RuntimeException("Match index " + matchIndex
					+ " out of range (" + m.groupCount() + "): " + m.pattern());

		handleExtraction(pieces[0], pieces[1], m.group(matchIndex));
	}

	/**
	 * Finds the method inside of which the text matched by the given
	 * Matcher is contained.
	 *
	 * This is not very intelligent, but for our purposes is perfect.
	 * It only recognizes a method as being declared with the pattern
	 * "^\\s*public.*method\\d+.*"; it will not match private or API
	 * methods, but that should never be needed.
	 *
	 * @param sf The SourceFile to search
	 * @param m The matcher used
	 * @return The name of the method, or null if not found
	 */
	public String findContainerMethod(SourceFile sf, Matcher m) {
		String[] lines = sf.getLines();
		int matchLine = sf.getContents().substring(0, m.start()).split("\n").length;

		for (int i = matchLine - 1; i >= 0; i--) {
			if (lines[i].matches("^\\s*public.*method\\d+.*"))
				return lines[i].substring(lines[i].indexOf("method"), lines[i].indexOf("("));
		}

		return null;
	}

	/**
	 * Gets the internal variable HashMap, into which items are stored
	 * by the 'define' keyword.
	 *
	 * @return The variable mapping
	 */
	public HashMap<String, String> getVariables() {
		return variables;
	}

	/**
	 * Gets the value of a particular internal variable.
	 *
	 * @return The variable's value, or null if not found.
	 */
	public String getVariable(String name) {
		return variables.get(name);
	}

	/**
	 * Sets the internal variable HashMap.
	 *
	 * @param variables The HashMap to set the internal variable map to
	 */
	public void setVariables(HashMap<String, String> variables) {
		this.variables = variables;
	}

	/**
	 * Sets the value of a particular internal variable.
	 *
	 * @param name The name of the variable to set
	 * @param val The value to give the variable
	 */
	public void setVariable(String name, String val) {
		variables.put(name, val);
	}

	/**
	 * Gets the HashMap<String, SourceFile> used to keep track
	 * of sources which have already been read.
	 *
	 * @return The full internal HashMap of SourceFiles
	 */
	public HashMap<String, SourceFile> getSources() {
		return sources;
	}

	/**
	 * Gets the SourceFile for the given filename.
	 *
	 * @param name The name of the file to get
	 * @return The corresponding SourceFile, or null if unavailable
	 */
	public SourceFile getSource(String name) {
		return sources.get(name);
	}

	/**
	 * Gets all of the currently active SourceFiles.
	 *
	 * This is only necessary for the special case of setting
	 * the current file to '*', in which case <code>getAvailableFiles()</code>
	 * is used to get a list of files in the input directory, and
	 * if <code>isClassRenamed(filename)</code> returns false, the
	 * file is add to the list of SourceFiles to be returned.
	 *
	 * Otherwise, the array contains only the current file's SourceFile.
	 *
	 * @return The SourceFiles that should be used 
	 * @see #getAvailableFiles()
	 * @see #isClassRenamed(String)
	 */
	public SourceFile[] getCurrentSources() {
		ArrayList<SourceFile> sfs = new ArrayList<SourceFile>();

		if (currentFile.equals("*")) {
			for (File f : getAvailableFiles()) {
				String className = f.getName().replace(".java", "");
				if (!isClassRenamed(className))
					sfs.add(new SourceFile(f));
			}
		} else {
			SourceFile sf = sources.get(currentFile);
			sfs.add(sf);
		}

		return sfs.toArray(new SourceFile[0]);
	}

	/**
	 * Sets the internal sources HashMap.
	 *
	 * @param sources The HashMap to set the internal map to.
	 */
	public void setSources(HashMap<String, SourceFile> sources) {
		this.sources = sources;
	}

	/**
	 * Sets a single SourceFile's value.
	 *
	 * @param name The name of the file
	 * @param sf The SourceFile to use
	 */
	public void setSource(String name, SourceFile sf) {
		sources.put(name, sf);
	}

	/**
	 * Gets the name of the currently active file.
	 *
	 * This is not necessarily a readable file; it may be '*',
	 * or nonexistent.
	 *
	 * @return The name of the currently active file
	 */
	public String getCurrentFile() {
		return currentFile;
	}

	/**
	 * Sets the current file, creating a source entry for it
	 * if necessary.
	 *
	 * If name is '*', the value is assigned literally, and should
	 * be handled elsewhere.
	 *
	 * @param name The name of the file to change to
	 */
	private void setCurrentFile(String name) {
		currentFile = name;

		if (!currentFile.equals("*") && !sources.containsKey(currentFile)) {
			SourceFile sf = new SourceFile(new File(inDir, currentFile));
			sources.put(currentFile, sf);
		}
	}

	/**
	 * Returns a line of text with all %name% variables expanded.
	 *
	 * Uses <code>findVariable()</code> to fetch the value.
	 *
	 * @param line The line of text to evaluate
	 * @return The resulting line of text
	 * @see #findVariable(String)
	 */
	public String expandVariables(String line) {
		String tmp = new String(line);

		do {
			Matcher m = varPattern.matcher(tmp);

			if (!m.find())
				break;

			String var = tmp.substring(m.start() + 1, m.end() - 1);
			String val = findVariable(var);

			if (val != null)
				tmp = tmp.replaceAll("%" + var + "%", val);
			else
				throw new RuntimeException("Unknown variable \"" + var + "\" in line: " + line);
		} while (true);

		return tmp;
	}

	/**
	 * Finds the value of a stored variable; fairly useless as is, as
	 * it is nothing more than a differently named getVariable(String).
	 *
	 * Should normally be overridden by subclasses.
	 *
	 * @param var The variable whose value should be fetched
	 * @return The value of the variable, or null if not found
	 */
	public String findVariable(String var) {
		return variables.get(var);
	}

	/**
	 * Gets the list of available .java files in the input directory.
	 *
	 * This is not recursive, which makes signlink.java off limits to us.
	 *
	 * @return The available .java files
	 * @see File#listFiles(FilenameFilter)
	 */
	public File[] getAvailableFiles() {
		return inDir.listFiles(new FilenameFilter() {
			public boolean accept(File d, String name) { return name.endsWith(".java"); }
		});
	}

	/**
	 * Implemented by the parser, serves no purpose here.
	 *
	 * In the parser, checks if the given class name has already
	 * been renamed.
	 *
	 * @param name The class to check
	 * @return Always false in this implementation.
	 */
	public boolean isClassRenamed(String name) {
		return false;
	}

	/**
	 * Implemented by the massager, serves no purpose here.
	 *
	 * In the massager, adds the given declaration to a file.
	 *
	 * @param val The line of text to add as a declaration.
	 */
	public void addDeclaration(String val) { }

	/**
	 * Takes the three given values from an extraction and does whatever
	 * is needed.
	 *
	 * @param type Corresponds to the keyword in an extraction instruction (class, constant)
	 * @param name The desired resulting name of 'matched'
	 * @param matched The text matched
	 * @see #handleLine(String, Matcher)
	 * @see #inMethod(String, BufferedReader, boolean)
	 */
	public abstract void handleExtraction(String type, String name, String matched);

	/**
	 * Saves the data however it should be done.
	 */
	public abstract void dump();

}
