import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.saev.deob.Application;

public class Renamer {
	private static final String DEFAULT_OUTFILE = "./renamed.jar";

	private Application app;
	private File confFile;
	private String outFile;

	private static boolean verbose;

	private ArrayList<JavaClass> finalClasses;
	private HashMap<String, String> classes;
	private HashMap<String, String> constants;

	private static int renameCount;

	public Renamer(String confFile, String inFile) {
		this(confFile, inFile, DEFAULT_OUTFILE);
	}

	public Renamer(String confFile, String inputFile, String outFile) {
		this.outFile = outFile;
		app = new Application(inputFile);

		finalClasses = new ArrayList<JavaClass>();
		classes = new HashMap<String, String>();
		constants = new HashMap<String, String>();

		parseConfig(new File(confFile));
	}

	public static void main(String[] args) {
		Renamer r;
		long startTime = System.currentTimeMillis();

		if (args.length < 2)
			fatal("Usage: Renamer configFile inputJar [outputJar]");

		verbose = System.getenv("VERBOSE") != null;
		if (args.length > 2)
			r = new Renamer(args[0], args[1], args[2]);
		else
			r = new Renamer(args[0], args[1]);

		r.rename();
		r.save();
		System.out.println(renameCount + " constant entries renamed in "
				+ (System.currentTimeMillis() - startTime) + "ms");
	}

	// doesn't bother with interfaces for the however minor speed
	// consideration, since it's not something required at this
	// point.
	public void rename() {
		debug("-- Beginning renaming process --");
		for (ClassGen cg : app.getClasses()) {
			String className = cg.getClassName();
			String superName = cg.getSuperclassName();

			if (classes.containsKey(className))
				cg.setClassName(classes.get(className));
			if (classes.containsKey(superName))
				cg.setSuperclassName(classes.get(superName));

			debug("  - In class: " + className + " = " + cg.getClassName());

			JavaClass jc = renameJavaClass(cg.getJavaClass());
			finalClasses.add(jc);
		}
	}

	// Application has no deleteClass ability, so we instead create a new
	// instance, add all the new JavaClass objects we've produced, then
	// dump the jarfile
	public void save() {
		Application finalApp = new Application();

		for (JavaClass jc : finalClasses)
			finalApp.setJavaClass(jc);

		finalApp.dumpJar(outFile);
	}

	public JavaClass renameJavaClass(JavaClass jc) {
		ConstantPool cp = jc.getConstantPool();
		Constant[] cons = cp.getConstantPool();

		for (int i = 0; i < cons.length; i++) {
			if (!(cons[i] instanceof ConstantUtf8))
				continue;

			ConstantUtf8 cutf = (ConstantUtf8)cons[i];

			for (Map.Entry<String, String> entry : constants.entrySet()) {
				String oldName = entry.getKey();
				String newName = entry.getValue();
				String val = cutf.getBytes();

				if (val.contains(oldName) && shouldAlter(oldName, val)) {
					debug("    - s/" + oldName + "(?!\\d)" + "/" + newName + "/: " + val);
					cutf.setBytes(val.replaceAll(oldName + "(?!\\d)", newName));
					cons[i] = cutf;
					++renameCount;
				}
			}
		}

		cp.setConstantPool(cons);
		jc.setConstantPool(cp);
		return jc;
	}

	// determines if we should alter the constant entry
	public boolean shouldAlter(String needle, String haystack) {
		Pattern p = Pattern.compile(".*?(a" + needle
				+ "(?!_Sub|\\d)(?:(?:Array)*|_)+\\d+).*?");
		Matcher m = p.matcher(haystack);

		if (m.find())
			return !constants.containsKey(m.group(1));

		return Pattern.matches(".*?" + needle + "(?!_Sub|\\d).*?", haystack);
	}

	// explicitly relies on the format dumped by my parser,
	// namely, ignores /^\[.*\]$/ and loads classes, then constants
	// after an empty line
	public void parseConfig(File f) {
		try {
			BufferedReader reader = new BufferedReader(new FileReader(f));
			boolean inClasses = true;
			String line;

			while ((line = reader.readLine()) != null) {
				if (line.trim().length() == 0) {
					inClasses = false;
					continue;
				}

				if (line.trim().matches("^\\[.*\\]$"))
					continue;

				String[] pieces = line.split("=");
				if (pieces.length != 2)
					fatal("Invalid input line: " + line);
				if (pieces[0].indexOf('.') != -1)
					pieces[0] = pieces[0].substring(line.lastIndexOf('.') + 1);

				// load in opposite order from how parser stores it.
				if (inClasses)
					classes.put(pieces[1], pieces[0]);
				constants.put(pieces[1], pieces[0]);
			}
		} catch (Exception e) { fatal("Exception parsing config file: " + e); }
	}

	public static void debug(String s) {
		if (verbose)
			System.out.println(s);
	}

	public static void fatal(String err) {
		System.err.println(err);
		System.exit(1);
	}
}
